import shutil
import tensorflow as tf

from . import losses
from . import metrics
from . import modes
from .train import TrainingHook, NotTrainingHook, optimizer as create_optimizer
from .utils import call_fn, logger, dataset


class Model():

    def __init__(self, model_fn, **kwargs):
        self._model_fn = model_fn
        self._estimator_kwargs = kwargs
        self._estimator = None
        self._hooks = self._get_hooks()

    def compile(self, **kwargs):
        self._estimator_kwargs.update(kwargs)
        self._estimator = self._create_estimator(model_fn=self._model_fn, **self._estimator_kwargs)

    def train(self, x, y=None, epochs=1, batch_size=128, shuffle=True, **kwargs):
        input_fn = dataset(x=x,
                           y=y,
                           epochs=epochs,
                           batch_size=batch_size,
                           shuffle=shuffle)
        self._update_kwargs(modes.TRAIN, kwargs)
        self.estimator.train(input_fn=input_fn, **kwargs)

    def evaluate(self, x, y=None, batch_size=128, **kwargs):
        input_fn = dataset(x=x,
                           y=y,
                           batch_size=batch_size,
                           shuffle=False)
        self._update_kwargs(modes.EVAL, kwargs)
        return self.estimator.evaluate(input_fn=input_fn, **kwargs)

    def predict(self, x, batch_size=128, **kwargs):
        input_fn = dataset(x=x,
                           batch_size=batch_size,
                           shuffle=False)
        self._update_kwargs(modes.PREDICT, kwargs)
        return self.estimator.predict(input_fn=input_fn, **kwargs)

    __call__ = predict

    @property
    def estimator(self):
        if self._estimator is None:
            self.compile()
        return self._estimator

    def _create_estimator(self, model_fn, directory=None, model_dir=None, params=None, **kwargs):
        defaults = self._defaults()
        if model_dir is None:
            model_dir = directory
        if model_dir is None:
            model_dir = defaults.get('model_dir')
            shutil.rmtree(model_dir, ignore_errors=True)
            logger.warn('Using temporary folder as model directory: {}'.format(model_dir))
        params = defaults.get('params', {}) if params is None else params
        model_fn = self._create_model_fn(model_fn)
        return tf.estimator.Estimator(model_fn=model_fn, model_dir=model_dir, params=params, **kwargs)

    def _defaults(self):
        return {
            'model_dir': '/tmp/nn_model_dir',
        }

    def _create_model_fn(self, model_fn):

        def fn(features, labels, mode, params, config):
            if list(features.keys()) == ['x']:
                features = features['x']

            ret = call_fn(model_fn, features, labels=labels, mode=mode, params=params, config=config)
            if not isinstance(ret, tf.estimator.EstimatorSpec):
                if not isinstance(ret, dict):
                    ret = dict(outputs=ret)
                ret = spec(mode=mode, labels=labels, **ret)
            return ret

        return fn

    def _get_hooks(self):
        return {
            modes.TRAIN: [TrainingHook()],
            modes.EVAL: [NotTrainingHook()],
            modes.PREDICT: [NotTrainingHook()],
        }

    def _update_kwargs(self, mode, kwargs):
        hooks = self._hooks.get(mode, []) + kwargs.get('hooks', [])
        if hooks:
            kwargs['hooks'] = hooks


def model(model_fn=None, **kwargs):

    def decorator(model_fn):
        return Model(model_fn, **kwargs)

    return decorator(model_fn) if model_fn is not None else decorator


def spec(mode, labels=None, outputs=None, predictions=None, loss=None, optimizer=None, metrics=None, **kwargs):
    if predictions is None:
        predictions = outputs
    if mode == modes.PREDICT:
        return tf.estimator.EstimatorSpec(mode=mode, predictions=predictions, **kwargs)

    loss = create_loss_op(loss, labels, outputs)

    if mode == modes.TRAIN:
        train_op = create_train_op(optimizer, loss)
        return tf.estimator.EstimatorSpec(mode=mode, loss=loss, train_op=train_op, **kwargs)

    metrics = create_metric_ops(metrics, labels, outputs)
    return tf.estimator.EstimatorSpec(mode=mode, loss=loss, eval_metric_ops=metrics, **kwargs)


def create_loss_op(loss, labels, outputs):
    if isinstance(loss, tf.Operation):
        return loss
    if isinstance(loss, str):
        loss = getattr(losses, loss)
    return loss(labels, outputs)


def create_train_op(optimizer, loss):
    if isinstance(optimizer, tf.Operation):
        return optimizer
    global_step = tf.train.get_global_step()
    if isinstance(optimizer, tuple):
        optimizer = create_optimizer(*optimizer)
    if not isinstance(optimizer, tf.train.Optimizer) and callable(optimizer):
        return call_fn(optimizer, loss, global_step=global_step)
    return optimizer.minimize(loss=loss, global_step=global_step)


def create_metric_ops(metrics, labels, outputs):
    metrics = metrics or {}
    if isinstance(metrics, list):
        metrics = {getattr(metric, '__name__', metric): metric for metric in metrics}
    return {name: create_metric_op(metric, labels, outputs) for name, metric in metrics.items()}


def create_metric_op(metric, labels, outputs):
    if isinstance(metric, tf.Operation):
        return metric
    if isinstance(metric, str):
        metric = getattr(metrics, metric)
    return metric(labels, outputs)



import tensorflow as tf

relu = tf.nn.relu
relu6 = tf.nn.relu6
crelu = tf.nn.crelu
elu = tf.nn.elu
selu = tf.nn.selu
softplus = tf.nn.softplus
softsign = tf.nn.softsign
softmax = tf.nn.softmax
sigmoid = tf.nn.sigmoid
tanh = tf.nn.tanh



import tensorflow as tf

gradientdescent = GradientDescent = tf.train.GradientDescentOptimizer
adadelta = Adadelta = tf.train.AdadeltaOptimizer
adagrad = Adagrad = tf.train.AdagradOptimizer
adagradda = AdagradDA = tf.train.AdagradDAOptimizer
momentum = Momentum = tf.train.MomentumOptimizer
adam = Adam = tf.train.AdamOptimizer
ftrl = Ftrl = tf.train.FtrlOptimizer
proximalgradientdescent = ProximalGradientDescent = tf.train.ProximalGradientDescentOptimizer
proximaladagrad = ProximalAdagrad = tf.train.ProximalAdagradOptimizer
rmsprop = RMSProp = tf.train.RMSPropOptimizer



from tensorflow.python.layers.layers import *
from .sparse import *
from .recurrent import *

import tensorflow as tf

from .. import activations
from ..train import training as is_training


def _uses_activation(Layer, attributes=None):
    if attributes is None:
        attributes = ['activation']

    def init(self, *args, **kwargs):
        super(Layer, self).__init__(*args, **kwargs)
        for name in attributes:
            if hasattr(self, name):
                activation = getattr(self, name)
                if isinstance(activation, str):
                    activation = getattr(activations, activation)
                    setattr(self, name, activation)

    Layer.__init__ = init
    return Layer


def _uses_training(Layer):

    def call(self, inputs, training=None, **kwargs):
        if training is None:
            training = is_training()
        return super(Layer, self).call(inputs, training=training, **kwargs)

    Layer.call = call
    return Layer


# Core

@_uses_activation
class Dense(tf.layers.Dense):
    pass


# Convolutional

@_uses_activation
class Conv1D(tf.layers.Conv1D):
    pass


@_uses_activation
class Conv2D(tf.layers.Conv2D):
    pass


@_uses_activation
class Conv2DTranspose(tf.layers.Conv2DTranspose):
    pass


@_uses_activation
class Conv3D(tf.layers.Conv3D):
    pass


@_uses_activation
class Conv3DTranspose(tf.layers.Conv3DTranspose):
    pass


@_uses_activation
class SeparableConv1D(tf.layers.SeparableConv1D):
    pass


@_uses_activation
class SeparableConv2D(tf.layers.SeparableConv2D):
    pass


# Training

@_uses_training
class Dropout(tf.layers.Dropout):
    pass


@_uses_training
class BatchNormalization(tf.layers.BatchNormalization):
    pass


# Others

def Reshape(shape, **keywords):
    shape = [-1] + shape  # preserve the batch axis (axis 0)
    return lambda tensor: tf.reshape(tensor, shape, **keywords)


def Sequence(layers):

    def layer(x):
        for l in layers:
            x = l(x)
        return x

    return layer


def concatenate(tensors):
    return tf.concat(tensors, 1)

__all__ = [
    'AveragePooling1D',
    'AveragePooling2D',
    'AveragePooling3D',
    'BatchNormalization',
    'Conv1D',
    'Conv2D',
    'Conv2DTranspose',
    'Conv3D',
    'Conv3DTranspose',
    'Dense',
    'Dropout',
    'Flatten',
    'InputSpec',
    'Layer',
    'MaxPooling1D',
    'MaxPooling2D',
    'MaxPooling3D',
    'SeparableConv1D',
    'SeparableConv2D',

    # Recurrent
    'RNN',
    'BasicLSTMCell',
    'BasicRNNCell',
    'DeviceWrapper',
    'DropoutWrapper',
    'GRUCell',
    'LSTMCell',
    'LSTMStateTuple',
    'MultiRNNCell',
    'RNNCell',
    'ResidualWrapper',

    # Sparse
    'Embedding',
]



import tensorflow as tf


class Embedding(tf.layers.Layer):

    def __init__(self, input_dim, output_dim,
                 embeddings_initializer=None,
                 embeddings_regularizer=None,
                 embeddings_constraint=None,
                 dtype=tf.float32,
                 **kwargs):
        super(Embedding, self).__init__(dtype=dtype, **kwargs)
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.embeddings_initializer = embeddings_initializer
        self.embeddings_regularizer = embeddings_regularizer
        self.embeddings_constraint = embeddings_constraint

    def build(self, input_shape):
        self.embeddings = self.add_variable(
            name='embeddings',
            shape=[self.input_dim, self.output_dim],
            initializer=self.embeddings_initializer,
            regularizer=self.embeddings_regularizer,
            constraint=self.embeddings_constraint,
            dtype=self.dtype,
            trainable=True)
        self.built = True

    def call(self, inputs):
        if inputs.dtype != 'int32':
            inputs = tf.cast(inputs, 'int32')
        return tf.nn.embedding_lookup(self.embeddings, inputs)

    def compute_output_shape(self, input_shape):
        return tf.TensorShape(input_shape).concatenate(self.output_dim)



from tensorflow.python.ops.rnn_cell_impl import *
import tensorflow as tf

from .. import sequence


def get_last_output(outputs, sequence_length):
    if sequence_length is not None:
        outputs = sequence.batch_gather_nd(outputs, sequence_length - 1)
    else:
        outputs = outputs[:, -1, :]
    return outputs


class RNN(tf.layers.Layer):

    def __init__(self, cell, cell_b=None, return_sequences=False, return_state=False, merge_mode='concat', **kwargs):
        self.cell = cell
        self.cell_b = cell_b
        self.return_sequences = return_sequences
        self.return_state = return_state
        self.merge_mode = merge_mode
        self._trainable = True
        self._bidirectional = cell_b is not None
        super(RNN, self).__init__(**kwargs)

    def call(self, inputs, sequence_length=None, **kwargs):
        kwargs['inputs'] = inputs
        kwargs['sequence_length'] = sequence_length
        kwargs['dtype'] = kwargs.get('dtype', tf.float32)

        outputs, state = (tf.nn.bidirectional_dynamic_rnn(self.cell, self.cell_b, **kwargs)
                          if self._bidirectional else tf.nn.dynamic_rnn(self.cell, **kwargs))

        if not self.return_sequences:
            if self._bidirectional:
                outputs_fw, outputs_bw = outputs
                outputs_fw = get_last_output(outputs_fw, sequence_length)
                outputs_bw = outputs_bw[:, 0, :]
                outputs = outputs_fw, outputs_bw
            else:
                outputs = get_last_output(outputs, sequence_length)

        if self._bidirectional:
            if self.merge_mode is 'concat':
                outputs = tf.concat(outputs, axis=-1)

        if self.return_state:
            return outputs, state
        else:
            return outputs

    @property
    def trainable(self):
        return self._trainable

    @trainable.setter
    def trainable(self, value):
        self._trainable = value
        self.cell.trainable = value
        if self.cell_b is not None:
            self.cell_b.trainable = value
